/*
 * xflag - an X client that displays your favorite flag.
 *
 * Copyleft (!c) 1995 Simon Yeh <syeh@csfcorp.com>
 *
 * Permission to use, copy, modify, distribute, and/or even make profit (if
 * you could) out of, this software is hereby granted without fee. 
 *
 * IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY, IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR 
 * CONSEQUENTIAL DEMAGES OR POLITICAL HARASSMENT/PERSECUTION WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA, PROFITS OR INDIVIDUAL FREEDOM,
 * WHEREAS IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <X11/Xos.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

#define RESENT		0x1
#define ANIMATE		0x2
#define CYCLE		0x4
#define TwoPi          	(64*360)
#define Pi             	(32*360)
#define FASTTICK     	(90)
#define NORMALTICK     	(360)
#define SLOWTICK     	(720)
#define MIN(a, b)      	(((b) < (a)) ? (b) : (a))

#define	oops(msg)	{ fprintf(stderr, "xflag: %s\n", msg); exit(-1); }
#define IsAnniversary(m,d,t)    ((t->tm_mon+1 == (m)) && (t->tm_mday == (d)))

#define deltaRGB(d, c0, c1, t) {\
		(d).red = (float)(((c1).red - (c0).red)/t); \
		(d).green = (float)(((c1).green - (c0).green)/t); \
		(d).blue = (float)(((c1).blue - (c0).blue)/t); \
	}

Pixmap base_pmp, sign_pmp;
Dimension width, height, width1, height1, sx, sy;

String red = "Red";
String blue = "Blue";
String white = "White";
String black = "Black";
String green = "Green";
String yellow = "Yellow";
String basefg, basebg, signfg, signbg;

enum FlagId { 
	ROC, PRC, USA, ROT, WHAT 
};

typedef struct {
	int		id;
	String		name,
			title;
	int		num_center,
			num_vertices,
			maxtick;
	unsigned int	flags;
	GC		basegc,
	    		signgc,
	    		anmgc;
} Flag;

Flag theflag, *flagp;

#define InitFlag(fp, i, n, t, c, v) { \
		(fp)->id = i; \
		(fp)->name = strdup(n); \
		(fp)->title = strdup(t); \
		(fp)->num_center = c; \
		(fp)->num_vertices = v; \
		(fp)->maxtick = NORMALTICK; \
		(fp)->basegc = (fp)->signgc = (fp)->anmgc = (GC)0; \
		(fp)->flags = 0x0; \
	}

typedef struct {
	Boolean animation;
	Boolean resent;
	Boolean cycle;
	char *name;
	char *speed;
} AppResStruct;

typedef struct {
	XtIntervalId 	id;
	unsigned long	looptime;
	XtAppContext 	app;
} TimeOutClientData;

typedef struct {
	float 	red,
		green,
		blue;
} rXColor;

/* command line options for resources */
static XrmOptionDescRec options[] = {
	 {"-a", "animation", XrmoptionNoArg, "True"},
	 {"-c", "cycle", XrmoptionNoArg, "True"},
	 {"-n", "name", XrmoptionSepArg, NULL},
	 {"-r", "resent", XrmoptionNoArg, "True"},
	 {"-s", "speed", XrmoptionSepArg, NULL}
};

static XtResource               resources[] =
{
#define offset(field) XtOffsetOf(AppResStruct, field)
        { "animation", "Animation", XmRBoolean, sizeof(Boolean),
	  offset(animation), XmRImmediate, (XtPointer)False },
        { "cycle", "Cycle", XmRBoolean, sizeof(Boolean),
	  offset(cycle), XmRImmediate, (XtPointer)False },
        { "name", "Name", XmRString, sizeof(String),
	  offset(name), XmRString, (XtPointer) "ROC" },
        { "resent", "Resent", XmRBoolean, sizeof(Boolean),
	  offset(resent), XmRImmediate, (XtPointer)False },
        { "speed", "Speed", XmRString, sizeof(String),
	  offset(speed), XmRString, (XtPointer) "normal" }
#undef offset
};

XPoint		orig, *Center, *Vertices;
AppResStruct	AppData, *AppDatap;
TimeOutClientData	tout_data;
Widget		toplevel, base;
Window		root_window;
GC		copygc;
rXColor		rvarColor, deltaColor;
XColor		varColor;
Colormap	cmap;
int 		scrn, depth;
short		r0, r1, r2;
double		deltar, curr;
Boolean		charge = FALSE;

Pixel		get_color();
void		set_color(), draw_flag(), draw_stripes(), 
		draw_symbol(), compute_data(),
		do_animation(), change_color(),
		draw(), redraw(), resize();

extern char 	*strdup();
extern double	rint();

main(argc, argv)
int argc;
char *argv[];
{
	Display *dpy;
	XtAppContext app;
	XGCValues gcv;
	int i;
	XtActionsRec actions;
	int fid;
	long clock;
	char buf[32];
	struct tm *tmp;

	if ((dpy = XOpenDisplay(0)) == NULL) {
		oops("XOpenDisplay()");
	}
	scrn = DefaultScreen(dpy);
	depth = DefaultDepth(dpy, scrn);
	if (depth <= 1) {
		oops("Must run on a color display");
	}

	XtSetLanguageProc (NULL, NULL, NULL);

	toplevel = XtAppInitialize (&app, "Xflag", options, XtNumber(options), 
	    &argc, argv, NULL, NULL, 0);

	if (argc > 2) usage(NULL);

	XtGetApplicationResources(toplevel, &AppData, resources,
	    XtNumber(resources), NULL, 0);

	AppDatap = &AppData;
	flagp = &theflag;
	switch (fid = name2id(AppDatap->name)) {
	case ROC:	
		InitFlag(flagp, fid, AppDatap->name, \
				"Blue Sky, White Sun and Red Land", 1, 24);
		break;
	case PRC:	
		InitFlag(flagp, fid, AppDatap->name, \
				"Red Star Over China", 5, 50);
		break;
	case USA:	
		InitFlag(flagp, fid, AppDatap->name, \
				"Stars and Stripes", 50, 500);
		break;
	case ROT:	
		InitFlag(flagp, fid, AppDatap->name, \
				"Republic of Taiwan", 8, 8);
		break;
	default:	
		usage("unknown flag");
	}
	if (AppDatap->animation) {
		if (flagp->id != ROT)
			oops("animation mode for this flag is not implemented");
		flagp->flags |= ANIMATE;
		if (AppDatap->cycle)
			flagp->flags |= CYCLE;
		if (isdigit(AppDatap->speed[0])) {
			if ((flagp->maxtick = atoi(AppDatap->speed)) <= 0)
				oops("speed must be greater than 0");
		} else {
			switch (AppDatap->speed[0]) {
			case 'f':
			case 'F': 
				flagp->maxtick = FASTTICK; 
				break;
			case 's':
			case 'S': 
				flagp->maxtick = SLOWTICK; 
				break;
			case 'n':
			case 'N': 
				flagp->maxtick = NORMALTICK; 
				break;
			default:  
				oops("illegal speed");
			}
		}
	}
	if (AppDatap->resent) {
		if (flagp->id != PRC)
			oops("resentment mode for this flag is not implemented");
		flagp->flags |= RESENT;
	}

	sprintf(buf, "%s Flag", flagp->name);
	XtVaSetValues(toplevel, XmNtitle, flagp->title, XmNiconName, buf, NULL);

	time( (time_t *) &clock);
	tmp =localtime( (time_t *) &clock);

	Center = (XPoint *)malloc(flagp->num_center * sizeof(XPoint));
	if (!Center) oops("malloc");

	Vertices = (XPoint *)malloc(flagp->num_vertices * sizeof(XPoint));
	if (!Vertices) oops("malloc");

	width = DisplayWidth(dpy, scrn)/3;
	height = DisplayHeight(dpy, scrn)/3;

	base = XtVaCreateManagedWidget ("base",
	    xmDrawingAreaWidgetClass, toplevel,
	    XmNwidth,        width,
	    XmNheight,       height,
	    NULL);
	XtAddCallback (base, XmNexposeCallback, redraw, NULL);
	XtAddCallback (base, XmNresizeCallback, resize, NULL);

	XtVaGetValues(base, XmNcolormap, &cmap, NULL);
	sx = sy = 0;

	basefg = signfg = white;
	basebg = red;
	signbg = blue;
	switch (flagp->id) {
	case USA:	
		width1 = width/2;
		height1 = (Dimension)(height*7/13);
		break;
	case ROT:	
		signfg = red;
		signbg = white;
		basebg = green;
		width1 = (Dimension)(width*3/5);
		height1 = height;
		sx = (Dimension)(width/5);
		break;
	case PRC:	
		if (IsAnniversary(4,5,tmp) || IsAnniversary(6,4,tmp))
			flagp->flags |= RESENT;
		if (flagp->flags & RESENT) {
			signbg = basebg = black;
			signfg = basefg = red;
		} else {
			signfg = basefg = yellow;
			signbg = red;
		}
	case ROC:	
		width1 = width/2;
		height1 = height/2;
		break;
	}

	root_window	= RootWindowOfScreen (XtScreen(base));
	dpy = XtDisplay(base);

	gcv.function = GXcopy;
	copygc = XCreateGC (dpy, root_window, GCFunction, &gcv);

	if (flagp->flags & ANIMATE) {
		unsigned long pixel, plane;
		XColor endC, startC;

		if (!XAllocColorCells (dpy, cmap, False, &plane, 0, &pixel, 1))
			oops("XAllocColorCells");
		XParseColor(dpy, cmap, basefg, &endC);
		XParseColor(dpy, cmap, basebg, &startC);
		deltaRGB(deltaColor, startC, endC, flagp->maxtick);

		curr = 0;
		gcv.foreground = get_color(base, signfg);
		flagp->anmgc = XCreateGC (dpy, root_window, GCForeground, &gcv);

		varColor.red = startC.red;
		varColor.green = startC.green;
		varColor.blue = startC.blue;
		rvarColor.red = varColor.red;
		rvarColor.green = varColor.green;
		rvarColor.blue = varColor.blue;
		varColor.flags = DoRed | DoGreen | DoBlue;
		gcv.foreground = varColor.pixel = pixel;
		XStoreColor (dpy, cmap, &varColor);

		tout_data.app = app;
		tout_data.looptime = 5L;
		XtAddEventHandler(base, ButtonPressMask, False,
		    do_animation, &tout_data);
	} else {
		gcv.foreground = get_color(base, basebg);
	}
	flagp->basegc = XCreateGC (dpy, root_window, GCForeground, &gcv);
	base_pmp = XCreatePixmap (dpy, root_window, width, height, depth);

	if (flagp->id != PRC) {
		gcv.foreground = get_color(base, signbg);
		flagp->signgc = XCreateGC (dpy, root_window, GCForeground, &gcv);
		sign_pmp = XCreatePixmap (dpy, root_window, width1, height1, depth);
	}

	compute_data(width1, height1);

	XtRealizeWidget (toplevel);

	draw_flag(0);

	XtAppMainLoop (app);
}

void draw_flag(sync)
int sync;
{
	Display *dpy = XtDisplay(base);

	XFillRectangle (dpy, base_pmp, flagp->basegc, 
		0, 0, width, height);
	if (sync)
		XFillRectangle (dpy, XtWindow(base), flagp->basegc, 
			0, 0, width, height);
	if (flagp->id == USA) draw_stripes(sync);

	if (flagp->id != PRC) {
		XFillRectangle (dpy, sign_pmp, flagp->signgc, 
			0, 0, width1, height1);
		if (sync)
			XFillRectangle (dpy, XtWindow(base), flagp->signgc, 
				sx, sy, width1, height1);
	}
	draw_symbol(sync);
}

void
redraw(w, client_data, call_data)
Widget    w;
XtPointer client_data;
XtPointer call_data;
{
	XmDrawingAreaCallbackStruct *cbs = 
	    (XmDrawingAreaCallbackStruct *) call_data;

	XCopyArea (cbs->event->xexpose.display, base_pmp, cbs->window, 
	    copygc, 0, 0, width, height, 0, 0);
	if (!sign_pmp) return;
	XCopyArea (cbs->event->xexpose.display, sign_pmp, cbs->window, 
	    copygc, 0, 0, width1, height1, sx, sy);
}

void
resize(w, client_data, call_data)
Widget    w;
XtPointer client_data;
XtPointer call_data;
{
	Dimension oldw, oldh;
	int sync = 0;
	XmDrawingAreaCallbackStruct *cbs = 
	    (XmDrawingAreaCallbackStruct *) call_data;

	oldw = width;
	oldh = height;

	XtVaGetValues(base, XmNwidth, &width, XmNheight, &height, NULL);
	if (flagp->id == ROT) {
		width1 = (Dimension)(width*3/5);
		height1 = height;
		sx = width/5;
	} else {
		width1 = width/2;
		height1 = (flagp->id == ROC) ? 
			height/2 : rint((double)(height*7/13));
	}
	compute_data(width1, height1);
	if (height > oldh || width > oldw) {
		XFreePixmap(XtDisplay(w), base_pmp);
		base_pmp = XCreatePixmap (XtDisplay(w), root_window,
		    width, height, depth);
		if (sign_pmp) {
			XFreePixmap(XtDisplay(w), sign_pmp);
			sign_pmp = XCreatePixmap (XtDisplay(w), root_window,
			    width1, height1, depth);
		}
	} else if (height <= oldh && width <= oldw){
		sync = 1;
	}
	draw_flag(sync);
}

Pixel get_color(widget, name)
Widget widget;
String name;
{
	Display *dpy = XtDisplay (widget);
	static XColor col, unused;

	if (!strcmp(name, "White"))
		return WhitePixelOfScreen (XtScreen (widget));
	if (!strcmp(name, "Black"))
		return BlackPixelOfScreen (XtScreen (widget));
	if (!XAllocNamedColor (dpy, cmap, name, &col, &unused)) {
		char buf[32];
		sprintf (buf, "Can't alloc %s", name);
		XtWarning (buf);
		return (Pixel)0;
	}
	return col.pixel;
}

void
set_color(widget, client_data, call_data)
Widget widget;
XtPointer client_data;
XtPointer call_data;
{
	Pixel pixel;

	pixel = get_color(widget, (String) client_data);
	XSetForeground (XtDisplay(widget), (GC) call_data, pixel);
}

int name2id(name)
String name;
{
	if (!strcasecmp(name, "ROC"))
		return ROC;
	if (!strcasecmp(name, "PRC"))
		return PRC;
	if (!strcasecmp(name, "USA"))
		return USA;
	if (!strcasecmp(name, "ROT"))
		return ROT;
	return WHAT;
}

void
getcord(center, pt, radius, d, l)
XPoint *center, *pt;
short radius, d, l;
{
	double degree;

	degree = ((double)d) * M_PI / l;
	pt->x = center->x + (short)rint((double)(radius * cos(degree)));
	pt->y = center->y + (short)rint((double)(radius * sin(degree)));
}

void
compute_data(w, h)
int w, h;
{
	Display *dpy = XtDisplay(base);
	XGCValues xgc;
	XPoint *xp, *ctp;
	short i, j, baseR;
	short r11, r22, k;

	baseR = MIN(w, h)/2;

	ctp = Center;
	xp = Vertices;
	if (flagp->id == ROC) {
		ctp->x = w/2;
		ctp->y = h/2;

		xgc.line_width = baseR/10;
		XChangeGC(dpy, flagp->signgc, GCLineWidth, &xgc);

		r0 = baseR * 0.3;
		r1 = r0 + xgc.line_width;
		r2 = baseR * 0.8;

		for (i = 0; i < flagp->num_vertices; i++, xp++)
			getcord(ctp, xp, (i % 2) ? r2 : r1, i+1, (short)12);
	} else if (flagp->id == PRC) {

		orig.x = w/2;
		orig.y = h/2;
		ctp->x = w/3;
		ctp->y = h/2;

		r0 = baseR * 0.7;
		r1 = baseR * 0.55;
		r2 = baseR * 0.18;
		r11 = r1 * 0.37;
		r22 = r2 * 0.38;

		for (i = 0; i < 10; i++, xp++)
			getcord(ctp, xp, (i % 2) ? r11 : r1, 2*i-1, (short)10);
		for (j = 1, ctp++; j < 5; j++, ctp++) {
			getcord(&orig, ctp, r0, (short)(2*j-5), (short)8);
			for (k = 0; k < 10; i++, xp++, k++)
				getcord(ctp, xp, (i % 2) ? r2 : r22, 2*i+1, (short)10);
		}
	} else if (flagp->id == ROT) {
		double deg, rf, rf2;

		deg = M_PI/8;
		orig.x = w/2;
		orig.y = h/2;
		r0 = baseR * 0.06;
		r1 = baseR * 0.4;
		rf = baseR * 0.4 / cos(deg);
		r11 = (short)rint(rf);
		rf2 = rf*sin(deg);
		r2 = (short)rint(rf2);

		for (j = 0; j < 8; j++, ctp++, xp++) {
			getcord(&orig, ctp, r0, 2*j, (short)8);
			getcord(ctp, xp, r11, 2*j, (short)8);
		}

		if (flagp->anmgc) {
			deltar = (double)(r0+rf+rf2) / (double)(flagp->maxtick);
			xgc.line_width = (short)ceil(3*deltar);
			XChangeGC(dpy, flagp->anmgc, GCLineWidth, &xgc);
		}
	} else {
		register short dx, dxx, dxy, dy;

		xgc.line_width = height/13;
		XChangeGC(dpy, flagp->basegc, GCLineWidth, &xgc);

		r0 = w/12;
		r1 = h/10;
		r2 = (MIN(w,h)/12) * 0.7;
		r22 = r2 * 0.4;
		for (j = 0, ctp++; j < 50; j++, ctp++) {
			dx = j % 11;
			dxx = dx % 6;
			dxy = dx / 6;
			dy = j / 11;
			ctp->x = (dxx*2 + dxy + 1)*r0;
			ctp->y = (dy*2 + dxy + 1)*r1;
			for (k = 0; k < 10; i++, xp++, k++)
				getcord(ctp, xp, (i % 2) ? r2 : r22, 2*i+1, (short)10);
		}
	}
}

void
draw_symbol(sync)
int sync;
{
	Display *dpy = XtDisplay(base);
	Pixmap pmp = sign_pmp;
	GC gc = flagp->signgc;
	XPoint *vp;
	int i, num_v;

	switch (flagp->id) {
	case ROC: 
		num_v = 24;		
		break;
	case PRC: 
		pmp = base_pmp;	
		gc = flagp->basegc;
	case USA: 
		num_v = 10;		
		break;
	case ROT: 
		num_v = 1;		
		break;
	}

	set_color(base, signfg, gc);
	if (flagp->id == ROT) {
		XPoint *cp = Center;
		short scurr = (short)curr;

		vp = Vertices;
		if ((flagp->flags & ANIMATE) && (curr > 0.0)) {
			XDrawArc(dpy, sign_pmp, flagp->anmgc, 
			    orig.x-scurr, orig.y-scurr,
			    2*scurr, 2*scurr, 0, TwoPi);
			if (sync) {
				XDrawArc(dpy, XtWindow(base), flagp->anmgc, 
			    		orig.x-scurr+sx, orig.y-scurr+sy,
			    		2*scurr, 2*scurr, 0, TwoPi);
			}
		}
		for (i=0; i < flagp->num_center; i++, cp++, vp++) {
			XFillArc(dpy, sign_pmp, gc, 
			    vp->x-r2, vp->y-r2,
			    2*r2, 2*r2, 0, TwoPi);
			XFillArc(dpy, sign_pmp, gc, 
			    cp->x-r1, cp->y-r1,
			    2*r1, 2*r1, Pi*(1-2*i)/8, -Pi/4);
			if (sync) {
				XFillArc(dpy, XtWindow(base), gc, 
				    vp->x-r2+sx, vp->y-r2+sy,
				    2*r2, 2*r2, 0, TwoPi);
				XFillArc(dpy, XtWindow(base), gc, 
				    cp->x-r1+sx, cp->y-r1+sy,
				    2*r1, 2*r1, Pi*(1-2*i)/8, -Pi/4);
			}
		}
	} else {
		for (i = 0, vp = Vertices; i < flagp->num_center; 
		    i++, vp += num_v) {
			XFillPolygon(dpy, pmp, gc, vp, num_v,
			    Nonconvex, CoordModeOrigin);
			if (sync)
				XFillPolygon(dpy, XtWindow(base), gc, vp, num_v,
			    		Nonconvex, CoordModeOrigin);
		}
	}

	set_color(base, signbg, gc);
	if (flagp->id == ROC) {
		XDrawArc(dpy, pmp, gc, 
		    Center[0].x-r1, Center[0].y-r1,
		    2*r1, 2*r1, 0, TwoPi);
		if (sync) {
			XDrawArc(dpy, XtWindow(base), gc, 
		    		Center[0].x-r1+sx, Center[0].y-r1+sy,
		    		2*r1, 2*r1, 0, TwoPi);
		}
	}
}

void
draw_stripes(sync)
int sync;
{
	Display *dpy = XtDisplay(base);
	int i;
	double y, dy;

	dy = height/13;
	y = dy * 1.5 + 1;
	dy += dy;
	set_color(base, basefg, flagp->basegc);
	for (i = 0; i < 11; i += 2, y += dy) {
		XDrawLine(dpy, base_pmp, flagp->basegc, 
		    0, (int)y, width, (int)y);
		if (sync) {
			XDrawLine(dpy, XtWindow(base), flagp->basegc, 
			    0, (int)y, width, (int)y);
		}
	}
	set_color(base, basebg, flagp->basegc);
}

void change_color(client_data, id)
XtPointer client_data;
XtIntervalId *id;
{
	TimeOutClientData *data = (TimeOutClientData *)client_data;
	static int tick = 0;
	static rXColor *rcp = &rvarColor;
	static XColor *cp = &varColor;

	if (tick++ % flagp->maxtick == 0) {
		charge = !charge;
		set_color(base, charge ? signfg : signbg, flagp->anmgc);
		curr = charge ? 0.0 : flagp->maxtick * deltar;
	}
	if (charge) {
		rcp->red   += deltaColor.red;
		rcp->green += deltaColor.green;
		rcp->blue  += deltaColor.blue;
		curr += deltar;
	} else {
		rcp->red   -= deltaColor.red;
		rcp->green -= deltaColor.green;
		rcp->blue  -= deltaColor.blue;
		curr -= deltar;
	}
	cp->red   = (unsigned short)rcp->red;
	cp->green = (unsigned short)rcp->green;
	cp->blue  = (unsigned short)rcp->blue;

	draw_symbol(1);
	XStoreColor(XtDisplay(base), cmap, cp);

	if (((flagp->flags & CYCLE) == 0x0) && (tick % flagp->maxtick == 0)) {
		XtRemoveTimeOut(data->id);
		data->id = 0;
	} else {
		data->id = XtAppAddTimeOut(data->app, data->looptime, 
		    change_color, (XtPointer) data);
	}
}

/* ARGSUSED */
void
do_animation(w, data, ev)
Widget w;
XtPointer data;
XEvent *ev;
{
	TimeOutClientData *tdp = (TimeOutClientData *)data;

	if (tdp->id) {
		XtRemoveTimeOut(tdp->id);
		tdp->id = 0;
	} else {
		change_color(tdp, tdp->id);
	}
}

usage(s)
char *s;
{
	if (s) fprintf(stderr, "xflag: %s\n", s);

	fprintf(stderr, 
	  "Usage: xflag -n <PRC|ROC|ROT|USA> [-acr] [-s <fast|normal|slow>]\n");
	exit(1);
}
